#include "Benchmarker.h"
#include "Application.h"
#include "Logger.h"

#include "OgreCamera.h"
#include "OgreConfigOptionMap.h"
#include "OgreRenderSystem.h"
#include "OgreRenderWindow.h"
#include "OgreRoot.h"


Benchmarker::Benchmarker(Application* application, Ogre::Camera* camera):
	mApplication(application),
	mCamera(camera),
	mTimeInCurrentPhase(0),
	mCurrentPhase(-1),
	mNumPhases(application->getNumMaterials()),
	mFramesRendered(0),
	mMeasureStartTime(1), //1
	mMeasureEndTime(6), //6
	mCamStartPos(-0.2f,  0.5f,  0.1f),
	mCamEndPos	(-0.1f,  0.5f,  0.2f),
	mIsFinished(false),
	mResultText()
{
	//print render options to verify that stuff was set correctly
	Ogre::Root& root = Ogre::Root::getSingleton();
	Ogre::RenderWindow* renderWindow = root.getAutoCreatedWindow();
	mResultText << "\nBENCHMARK Volumetrics v" << mApplication->getVersionNumber() << "\nSystem information:";
	const Ogre::ConfigOptionMap& configOptions = root.getRenderSystem()->getConfigOptions();
	Ogre::ConfigOptionMap::const_iterator iter = configOptions.begin();
	Ogre::ConfigOptionMap::const_iterator end = configOptions.end();
	while (iter != end)
	{
		mResultText
			<< "\n  "
			<< iter->first
			<< ": "
			<< iter->second.currentValue;
		iter++;
	}

	mResultText << "\nResults:\n";

	//start at the first material
	mApplication->setCurrentMaterialIndex(0);

	//turn of all particles
	mApplication->setParticleSprayActive(false);

	//set the initial material
	switchToNextPhase();
}


Benchmarker::~Benchmarker()
{
	mApplication->setParticleSprayActive(true);
	if (!mIsFinished)
	{
		std::string notFinishedText = "BENCHMARKING ABORTED, NOT FINISHED!";
		Logger::log(notFinishedText);
		mResultText << "\n" << notFinishedText;
	}
	Logger::log(mResultText.str());
}


void Benchmarker::update(float time)
{
	mTimeInCurrentPhase += time;
	if (mTimeInCurrentPhase >= mMeasureEndTime)
	{
		switchToNextPhase();
	}
	if (mTimeInCurrentPhase >= mMeasureStartTime)
	{
		mFramesRendered++;
	}

	updateCamera();
}


bool Benchmarker::isFinished() const
{
	return mIsFinished;
}


void Benchmarker::updateCamera()
{
	float interpolator = mTimeInCurrentPhase / mMeasureEndTime;
	interpolator *= 6.2832f;
	Ogre::Vector3 rotationCentre(0, 0, 0);
	Ogre::Vector3 localCamPos(sin(interpolator), 0.3f, cos(interpolator));
	mCamera->setPosition(rotationCentre + localCamPos * 8);
	mCamera->lookAt(rotationCentre);
}


void Benchmarker::switchToNextPhase()
{
	if (mCurrentPhase != -1)
	{
		mResultText << mFramesRendered << "\n";
	}
	mResultText << "  ";

	mTimeInCurrentPhase = 0;
	mCurrentPhase++;
	mFramesRendered = 0;
	std::stringstream text;
	text
		<< "Volumetrics version "
		<< mApplication->getVersionNumber()
		<< "\nRunning benchmark"
		<< "\nProgress: phase "
		<< mCurrentPhase + 1
		<< " of "
		<< mNumPhases
		<< " ("
		<< mCurrentPhase * mMeasureEndTime
		<< "s of "
		<< mNumPhases * mMeasureEndTime
		<< "s elapsed)"
		<< "\nCurrently testing: ";

	/*
	Phases:
		0 to numMaterials
			once with every material
	*/
	if (mCurrentPhase < mNumPhases)
	{
		text << mApplication->getCurrentMaterialName();

		mApplication->setCurrentMaterialIndex(mCurrentPhase);

		mResultText
			<< "Different materials\t"
			<< mApplication->getCurrentMaterialName()
			<< "\t1\t";
	}
	else
	{
		mResultText << "Finished benchmarking!";
		mIsFinished = true;
	}

	mApplication->setOverlayTexts(text.str(), "");
}