#include "ExampleFrameListener.h"
#include "InputManager.h"
#include "Logger.h"

#include "OgreCamera.h"
#include "OgreException.h"
#include "OgreOverlay.h"
#include "OgreOverlayElement.h"
#include "OgreOverlayManager.h"
#include "OgreRenderWindow.h"
#include "OgreStringConverter.h"
#include "OgreVector3.h"


ExampleFrameListener::ExampleFrameListener(Ogre::RenderWindow* win, Ogre::Camera* cam):
	mRotateSpeed(36),
    mMoveSpeed(2),
    mCamera(cam),
    mWindow(win),
    mStatsOn(true),
	mNumScreenShots(0),
	mTimeUntilNextToggle(0),
    mSceneDetailIndex(0),
    mMoveScale(0.0f),
    mRotScale(0.0f),
	mTranslateVector(Ogre::Vector3::ZERO),
    mAniso(1),
    mFiltering(Ogre::TFO_BILINEAR),
	mCameraRotationAllowed(false)
{
    mDebugOverlay = Ogre::OverlayManager::getSingleton().getByName("Core/DebugOverlay");

    showDebugOverlay(true);
}


ExampleFrameListener::~ExampleFrameListener()
{
}


bool ExampleFrameListener::processUnbufferedKeyInput(const Ogre::FrameEvent& evt)
{
	InputManager* inputManager = InputManager::get();

    if (inputManager->isKeyDown(OIS::KC_A))
    {
        // Move camera left
        mTranslateVector.x = -mMoveScale;
    }

    if (inputManager->isKeyDown(OIS::KC_D))
    {
        // Move camera RIGHT
        mTranslateVector.x = mMoveScale;
    }

    /* Move camera forward by keypress. */
    if (inputManager->isKeyDown(OIS::KC_UP) || inputManager->isKeyDown(OIS::KC_W) )
    {
        mTranslateVector.z = -mMoveScale;
    }

    /* Move camera backward by keypress. */
    if (inputManager->isKeyDown(OIS::KC_DOWN) || inputManager->isKeyDown(OIS::KC_S) )
    {
        mTranslateVector.z = mMoveScale;
    }

    if (inputManager->isKeyDown(OIS::KC_PGUP))
    {
        // Move camera up
        mTranslateVector.y = mMoveScale;
    }

    if (inputManager->isKeyDown(OIS::KC_PGDOWN))
    {
        // Move camera down
        mTranslateVector.y = -mMoveScale;
    }

	if (mCameraRotationAllowed)
	{
		if (inputManager->isKeyDown(OIS::KC_RIGHT))
		{
			mCamera->yaw(-mRotScale);
		}
		
		if (inputManager->isKeyDown(OIS::KC_LEFT))
		{
			mCamera->yaw(mRotScale);
		}
	}

    if (inputManager->isKeyDown(OIS::KC_ESCAPE))
    {            
        return false;
    }

    if (inputManager->isKeyDown(OIS::KC_F) && mTimeUntilNextToggle <= 0)
    {
        mStatsOn = !mStatsOn;
        showDebugOverlay(mStatsOn);

        mTimeUntilNextToggle = 1;
    }

    if (inputManager->isKeyDown(OIS::KC_SYSRQ) && mTimeUntilNextToggle <= 0)
    {
		char tmp[20];
		sprintf(tmp, "screenshot_%d.png", ++mNumScreenShots);
        mWindow->writeContentsToFile(tmp);
        mTimeUntilNextToggle = 0.5;
    }
	
	if (inputManager->isKeyDown(OIS::KC_R) && mTimeUntilNextToggle <=0)
	{
		mSceneDetailIndex = (mSceneDetailIndex + 1) % 2;
		switch(mSceneDetailIndex)
		{
			case 0:
				mCamera->setPolygonMode(Ogre::PM_SOLID);
				break;
			case 1:
				mCamera->setPolygonMode(Ogre::PM_WIREFRAME);
				break;
		}
		mTimeUntilNextToggle = 0.5;
	}

    // Return true to continue rendering
    return true;
}


bool ExampleFrameListener::processUnbufferedMouseInput(const Ogre::FrameEvent& evt)
{
	InputManager* inputManager = InputManager::get();

	float relX = inputManager->getMouseRelativeX();
	float relY = inputManager->getMouseRelativeY();

	//If the second mouse button is pressed, then the mouse movement results in 
	//sliding the camera, otherwise we rotate.

	//This next line contains the command InputManager::MB_RIGHT, but this mysteriously
	//gives compiler errors. No clue why, so mouse sliding is disabled for the moment.
	if (false)//inputManager->isMouseButtonDown(InputManager::MB_RIGHT))
	{
		mTranslateVector.x += relX * 0.013f;
		mTranslateVector.y -= relY * 0.013f;
	}
	else
	{
		mRotX = Ogre::Degree(-relX * 0.13f);
		mRotY = Ogre::Degree(-relY * 0.13f);
	}

	return true;
}


void ExampleFrameListener::moveCamera()
{
    // Make all the changes to the camera
    // Note that YAW direction is around a fixed axis (freelook style) rather than a natural YAW (e.g. airplane)
	if (mCameraRotationAllowed)
	{
		mCamera->yaw(mRotX);
		mCamera->pitch(mRotY);
	}
    mCamera->moveRelative(mTranslateVector);
}


void ExampleFrameListener::showDebugOverlay(bool show)
{
    if (mDebugOverlay)
    {
        if (show)
        {
            mDebugOverlay->show();
        }
        else
        {
            mDebugOverlay->hide();
        }
    }
}


// Override frameStarted event to process that (don't care about frameEnded)
bool ExampleFrameListener::frameStarted(const Ogre::FrameEvent& evt)
{
    if (mWindow->isClosed())
	{
        return false;
	}

	InputManager::get()->capture();

	// one of the input modes is immediate, so setup what is needed for immediate mouse/key movement
	if (mTimeUntilNextToggle >= 0) 
	{
		mTimeUntilNextToggle -= evt.timeSinceLastFrame;
	}

	mMoveScale = mMoveSpeed * evt.timeSinceLastFrame;
	mRotScale = mRotateSpeed * evt.timeSinceLastFrame;
	mRotX = 0;
    mRotY = 0;
	mTranslateVector = Ogre::Vector3::ZERO;

	if (!processUnbufferedKeyInput(evt))
	{
		return false;
	}

	if (!processUnbufferedMouseInput(evt))
	{
		return false;
	}

	moveCamera();

	return true;
}


bool ExampleFrameListener::frameEnded(const Ogre::FrameEvent& evt)
{
    updateStats();
    return true;
}


void ExampleFrameListener::setCameraRotationAllowed(bool allowed)
{
	mCameraRotationAllowed = allowed;
}


void ExampleFrameListener::updateStats()
{
    static Ogre::String currFps = "Current FPS: ";
    static Ogre::String avgFps = "Average FPS: ";
    static Ogre::String bestFps = "Best FPS: ";
    static Ogre::String worstFps = "Worst FPS: ";
    static Ogre::String tris = "Triangle Count: ";

    // update stats when necessary
    try
	{
        Ogre::OverlayElement* guiAvg = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/AverageFps");
        Ogre::OverlayElement* guiCurr = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/CurrFps");
        Ogre::OverlayElement* guiBest = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/BestFps");
        Ogre::OverlayElement* guiWorst = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/WorstFps");

        const Ogre::RenderTarget::FrameStats& stats = mWindow->getStatistics();

        guiAvg->setCaption(avgFps + Ogre::StringConverter::toString(stats.avgFPS));
        guiCurr->setCaption(currFps + Ogre::StringConverter::toString(stats.lastFPS));
        guiBest->setCaption(bestFps + Ogre::StringConverter::toString(stats.bestFPS)
            + " " + Ogre::StringConverter::toString(stats.bestFrameTime)+" ms");
        guiWorst->setCaption(worstFps + Ogre::StringConverter::toString(stats.worstFPS)
            + " " + Ogre::StringConverter::toString(stats.worstFrameTime)+" ms");

        Ogre::OverlayElement* guiTris = Ogre::OverlayManager::getSingleton().getOverlayElement("Core/NumTris");
        guiTris->setCaption(tris + Ogre::StringConverter::toString(stats.triangleCount));
    }
    catch (...)
    {
        // ignore
    }
}